package limax.endpoint;

import java.math.BigInteger;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

import limax.codec.Codec;
import limax.codec.HmacMD5;
import limax.codec.Octets;
import limax.defines.ErrorCodes;
import limax.defines.ErrorSource;
import limax.defines.VariantDefines;
import limax.endpoint.ViewContext.Type;
import limax.endpoint.providerendpoint.SyncViewToClients;
import limax.endpoint.switcherendpoint.CHandShake;
import limax.endpoint.switcherendpoint.OnlineAnnounce;
import limax.endpoint.switcherendpoint.PingAndKeepAlive;
import limax.endpoint.switcherendpoint.SHandShake;
import limax.endpoint.switcherendpoint.SessionKick;
import limax.endpoint.switcherendpoint.SessionLoginByToken;
import limax.net.ClientListener;
import limax.net.ClientManager;
import limax.net.Config;
import limax.net.Engine;
import limax.net.Listener;
import limax.net.Manager;
import limax.net.StateTransport;
import limax.net.SupportDispatch;
import limax.net.Transport;
import limax.util.Helper;
import limax.util.Trace;

final class EndpointManagerImpl implements EndpointManager, ClientListener, SupportDispatch {
	private static class ViewContextMap {
		private final Map<Short, Collection<AbstractViewContext>> map = new HashMap<Short, Collection<AbstractViewContext>>();

		void put(short pvid, AbstractViewContext vc) {
			Collection<AbstractViewContext> c = map.get(pvid);
			if (null == c)
				map.put(pvid, c = new ArrayList<AbstractViewContext>());
			c.add(vc);
		}

		void onSyncViewToClients(SyncViewToClients protocol) throws Exception {
			Collection<AbstractViewContext> c = map.get(protocol.providerid);
			if (c != null)
				for (AbstractViewContext vc : c)
					vc.onSyncViewToClients(protocol);
		}

		void clear() {
			for (Collection<AbstractViewContext> c : map.values())
				for (AbstractViewContext vc : c)
					try {
						vc.clear();
					} catch (Exception e) {
					}
			map.clear();
		}

		ViewContext getViewContext(short pvid, Type type) {
			Collection<AbstractViewContext> c = map.get(pvid);
			if (c != null)
				for (AbstractViewContext vc : c)
					if (type.equals(vc.getType()))
						return vc;
			return null;
		}
	}

	ClientManager manager;

	private final EndpointConfig config;
	private final EndpointListener listener;
	private final Map<Short, Byte> pvids;
	private final ScriptExchange scriptExchange;
	private final ViewContextMap viewContextMap = new ViewContextMap();

	public EndpointManagerImpl(EndpointConfig config, EndpointListener listener, Map<Short, Byte> pvids)
			throws Exception {
		this.config = config;
		this.listener = listener;
		this.pvids = pvids;
		this.manager = (ClientManager) Engine.add(config.getClientManagerConfig(), this, this);
		this.scriptExchange = config.getScriptEngineHandle() != null
				? new ScriptExchange(this, config.getScriptEngineHandle()) : null;
		if (config.isPingServerOnly())
			canretrylogin = false;
	}

	@Override
	public Transport getTransport() {
		return manager.getTransport();
	}

	@Override
	public void close() {
		manager.close();
	}

	@Override
	public void close(Transport transport) {
		manager.close(transport);
	}

	@Override
	public void dispatch(Runnable r, Object hit) {
		((SupportDispatch) manager).dispatch(r, hit);
	}

	@Override
	public Listener getListener() {
		return listener;
	}

	@Override
	public Manager getWrapperManager() {
		return null;
	}

	@Override
	public void onManagerInitialized(Manager manager, Config config) {
		if (LoginStatus.RETRY == loginstatus)
			loginstatus = LoginStatus.LOGINING;
		else
			listener.onManagerInitialized(this, this.config);
	}

	@Override
	public void onManagerUninitialized(Manager manager) {
		if (LoginStatus.RETRY == loginstatus) {
			Engine.getProtocolScheduler().execute(new Runnable() {
				@Override
				public void run() {
					try {
						EndpointManagerImpl.this.manager = (ClientManager) Engine.add(
								EndpointManagerImpl.this.config.getClientManagerConfig(), EndpointManagerImpl.this,
								EndpointManagerImpl.this);
					} catch (Exception e) {
						if (Trace.isErrorEnabled())
							Trace.error("session logined retry", e);
					}
				}
			});
		} else {
			if (LoginStatus.LOGINED_DONE == loginstatus) {
				try {
					keepalive.cancel();
					listener.onTransportRemoved(transportsaved);
				} catch (Exception e) {
					if (Trace.isErrorEnabled())
						Trace.error("listener.onTransportRemoved", e);
				}
			}
			if (LoginStatus.LOGINED_DONE == loginstatus || LoginStatus.LOGINED_NOTIFY == loginstatus) {
				Endpoint.clearDefaultEndpointManager(this);
				viewContextMap.clear();
				if (scriptExchange != null)
					scriptExchange.onUnload();
			}
			listener.onManagerUninitialized(this);
		}
	}

	@Override
	public void onTransportAdded(Transport transport) throws Exception {
		listener.onSocketConnected();
		if (config.isPingServerOnly()) {
			keepalive.startPingAndKeepAlive(transport);
		} else {
			dhRandom = Helper.makeDHRandom();
			new CHandShake((byte) config.getDHGroup(),
					Octets.wrap(Helper.generateDHResponse(config.getDHGroup(), dhRandom).toByteArray()))
							.send(transport);
		}
	}

	@Override
	public void onTransportRemoved(Transport transport) throws Exception {
		if (keepalive.isTimeout())
			onErrorOccured(ErrorSource.ENDPOINT, ErrorCodes.ENDPOINT_PING_TIMEOUT);
		if (LoginStatus.LOGINING == loginstatus) {
			if (!config.isPingServerOnly())
				listener.onAbort(transport);
		} else
			transportsaved = transport;
	}

	private void onErrorOccured(int source, int code) throws Exception {
		listener.onErrorOccured(source, code, null);
	}

	@Override
	public void onAbort(Transport transport) throws Exception {
		listener.onAbort(transport);
	}

	@Override
	public Config getConfig() {
		return config;
	}

	@Override
	public long getSessionId() {
		return sessionid;
	}

	@Override
	public long getAccountFlags() {
		return accountflags;
	}

	@Override
	public String toString() {
		return getClass().getName();
	}

	PortForward.ManagerImpl getPortForwardManager() {
		final PortForward.ManagerImpl pfm;
		synchronized (this) {
			if (null == portforward)
				portforward = new PortForward.ManagerImpl(this);
			pfm = portforward;
		}
		return pfm;
	}

	private class KeepAlive {
		private final static int PING_TIMEOUT = 5;
		private final static int KEEP_ALIVE_DELAY = 50;
		private final AtomicReference<Future<?>> refFuture = new AtomicReference<Future<?>>();
		private volatile boolean timeout = false;

		private void update(Future<?> newFuture) {
			Future<?> future = refFuture.getAndSet(newFuture);
			if (future != null)
				future.cancel(false);
		}

		public void cancel() {
			update(null);
		}

		public void startPingAndKeepAlive(Transport transport) throws Exception {
			new PingAndKeepAlive(System.currentTimeMillis()).send(transport);
			update(Engine.getProtocolScheduler().scheduleWithFixedDelay(new Runnable() {
				@Override
				public void run() {
					timeout = true;
					Transport transport = getTransport();
					if (transport != null)
						close(transport);
				}
			}, PING_TIMEOUT, Long.MAX_VALUE, TimeUnit.SECONDS));
		}

		public void process(PingAndKeepAlive p) {
			if (timeout)
				return;
			listener.onKeepAlived((int) (System.currentTimeMillis() - p.timestamp));
			if (config.isPingServerOnly())
				return;
			update(Engine.getProtocolScheduler().scheduleWithFixedDelay(new Runnable() {
				@Override
				public void run() {
					try {
						startPingAndKeepAlive(getTransport());
					} catch (Exception e) {
					}
				}
			}, KEEP_ALIVE_DELAY, Long.MAX_VALUE, TimeUnit.SECONDS));
		}

		public boolean isTimeout() {
			return timeout;
		}
	}

	private final KeepAlive keepalive = new KeepAlive();
	private volatile BigInteger dhRandom;
	private volatile PortForward.ManagerImpl portforward = null;
	private volatile Transport transportsaved = null;
	private volatile boolean canretrylogin = true;
	private volatile long sessionid = -1;
	private volatile long accountflags = 0;

	static enum LoginStatus {
		LOGINING, LOGINED_NOTIFY, LOGINED_DONE, RETRY,
	}

	private volatile LoginStatus loginstatus = LoginStatus.LOGINING;

	void process(SHandShake p) throws Exception {
		StateTransport transport = (StateTransport) p.getTransport();
		byte[] material = Helper.computeDHKey(config.getDHGroup(), new BigInteger(p.dh_data.getBytes()), dhRandom)
				.toByteArray();
		byte[] key = ((InetSocketAddress) transport.getPeerAddress()).getAddress().getAddress();
		int half = material.length / 2;
		HmacMD5 mac = new HmacMD5(Codec.Null(), key, 0, key.length);
		mac.update(material, 0, half);
		transport.setOutputSecurityCodec(mac.digest(), p.c2sneedcompress);
		mac = new HmacMD5(Codec.Null(), key, 0, key.length);
		mac.update(material, half, material.length - half);
		transport.setInputSecurityCodec(mac.digest(), p.s2cneedcompress);
		transport.setState(limax.endpoint.states.Endpoint.EndpointSessionLogin);
		listener.onKeyExchangeDone();
		SessionLoginByToken protocol = new SessionLoginByToken();
		protocol.username = config.getUserName();
		protocol.token = config.getToken();
		protocol.platflag = config.getPlatFlag();
		protocol.pvids.putAll(pvids);
		InetSocketAddress address = (InetSocketAddress) transport.getLocalAddress();
		protocol.report_ip.replace(address.getAddress().getAddress());
		protocol.report_port = (short) address.getPort();
		protocol.send(transport);
		transport.setState(config.getEndpointState());
	}

	void process(SessionKick p) throws Exception {
		onErrorOccured(ErrorSource.LIMAX, p.error);
		if (scriptExchange != null)
			scriptExchange.onClose(p.error);
	}

	void process(limax.endpoint.switcherendpoint.PortForward p) throws Exception {
		if (null != portforward)
			portforward.onProtocol(p);
	}

	void process(PingAndKeepAlive p) throws Exception {
		keepalive.process(p);
	}

	void process(OnlineAnnounce p) throws Exception {
		final Transport transport = p.getTransport();
		if (ErrorSource.LIMAX == p.errorSource && ErrorCodes.SUCCEED == p.errorCode) {
			sessionid = p.sid;
			accountflags = p.flags;
			createStaticViewContextImpls();
			parseVariantDefines(p.variantdefines);
			if (scriptExchange != null)
				scriptExchange.onLoad(p.scriptdefines);
			loginstatus = LoginStatus.LOGINED_NOTIFY;
			Endpoint.setDefaultEndpointManager(this);
			listener.onTransportAdded(transport);
			loginstatus = LoginStatus.LOGINED_DONE;
			keepalive.startPingAndKeepAlive(transport);
		} else {
			if (ErrorSource.LIMAX == p.errorSource && ErrorCodes.SWITCHER_SESSION_LOGINED == p.errorCode
					&& canretrylogin) {
				canretrylogin = false;
				loginstatus = LoginStatus.RETRY;
			} else {
				onErrorOccured(p.errorSource, p.errorCode);
			}
			close();
		}
	}

	private void parseVariantDefines(Map<Short, VariantDefines> variantdefines) {
		for (Map.Entry<Short, VariantDefines> e : variantdefines.entrySet()) {
			final Short pvid = e.getKey();
			viewContextMap.put(pvid, VariantViewContextImpl.createInstance(pvid, e.getValue(), this));
		}
	}

	private void createStaticViewContextImpls() {
		for (Map.Entry<Short, Map<Short, Class<? extends View>>> e : config.getStaticViewClasses().entrySet())
			viewContextMap.put(e.getKey(), new StaticViewContextImpl(e.getKey(), e.getValue(), this));

		{
			limax.endpoint.auanyviews.ViewManager vm = limax.endpoint.auanyviews.ViewManager.getInstance();
			viewContextMap.put(vm.getProviderId(),
					new StaticViewContextImpl(vm.getProviderId(), vm.getClasses(), this));
		}
	}

	void process(SyncViewToClients p) throws Exception {
		viewContextMap.onSyncViewToClients(p);
		if (scriptExchange != null)
			scriptExchange.onSyncViewToClients(p);
	}

	@Override
	public ViewContext getViewContext(short pvid, Type type) {
		return viewContextMap.getViewContext(pvid, type);
	}

}
