package limax.provider;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import limax.codec.Octets;
import limax.codec.StringStream;
import limax.provider.providerendpoint.SyncViewToClients;
import limax.providerendpoint.ViewVariableData;
import limax.util.Resource;
import limax.util.Trace;

abstract class AutoView extends View {

	private static class FieldUpdater {
		private static class Data {
			private final static Octets NullData = new Octets(0);
			private final byte field;
			private final Octets data;
			private final Octets dataremoved;

			Data(byte field, Octets data, Octets dataremoved) {
				this.field = field;
				this.data = data;
				this.dataremoved = dataremoved;
			}

			Data(byte field, Octets data) {
				this(field, data, NullData);
			}

			int estimateDataSize() {
				return data.size() + dataremoved.size() + 4;
			}

			ViewVariableData asViewVariableData(byte index) {
				return new ViewVariableData(index, field, data, dataremoved);
			}
		}

		private final List<Data> data = new ArrayList<>();
		private StringBuilder text = new StringBuilder();

		void update(byte fldindex, Octets flddata, String fldstr) {
			data.add(new Data(fldindex, flddata));
			text.append(fldstr);
		}

		void update(byte fldindex, Octets added, Octets removed, String fldstr) {
			data.add(new Data(fldindex, added, removed));
			text.append(fldstr);
		}

		void clear() {
			data.clear();
			text = new StringBuilder();
		}

		DataItem merge(DataItem item) {
			if (data.isEmpty())
				return item;
			int fullDataSize = item.data.data.size();
			int fieldsDataSize = data.stream().mapToInt(i -> i.estimateDataSize()).sum();
			int fullTextSize = item.text.length();
			int fieldsTextSize = text.length() + item.text.indexOf('?', 1) + 3;
			String string;
			if (fullTextSize <= fieldsTextSize) {
				if (fullDataSize <= fieldsDataSize)
					return item;
				string = item.text;
			} else
				string = item.text.substring(0, item.text.indexOf('?', 1) + 1) + "W" + text + ":";
			return new DataItem(item.data, item.text,
					fullDataSize <= fieldsDataSize ? Arrays.asList(new DataItem(item.data, string))
							: Stream.concat(
									data.stream().map(i -> new DataItem(i.asViewVariableData(item.data.index), "")),
									Stream.of(new DataItem(
											new ViewVariableData(item.data.index, (byte) -1, NullData, NullData),
											string)))
									.collect(Collectors.toList()));
		}
	}

	private final String[] touch;
	private SyncViewToClients protocol;
	private StringStream sstream;
	private final FieldUpdater fieldUpdater = new FieldUpdater();

	abstract Collection<Long> getSessionIds();

	abstract void assembleProtocolData(SyncViewToClients protocol, StringStream sstream);

	abstract byte getSyncDataType();

	protected AutoView(ViewContext context, ViewStub stub, Resource resource, String[] touch) {
		super(context, stub, resource, touch.length);
		this.touch = touch;
	}

	@Override
	String getTouch(byte varindex) {
		return touch[varindex];
	}

	void delaySchedule(Runnable task) {
		getViewContext().delaySchedule(task);
	}

	int getInstanceIndex() {
		return 0;
	}

	protected final void updateField(byte fldindex, Octets flddata, String fldstr) {
		hashSchedule(() -> fieldUpdater.update(fldindex, flddata, fldstr));
	}

	protected final void updateField(byte fldindex, Octets added, Octets removed, String updatestr) {
		hashSchedule(() -> fieldUpdater.update(fldindex, added, removed, updatestr));
	}

	@Override
	DataItem syncToClient(byte varindex, Octets vardata, String strdata) {
		try {
			DataItem item = super.syncToClient(varindex, vardata, strdata);
			if (item == null)
				return null;
			if (item.data.index < 0 || item.data.data.size() == 0)
				return syncToClient(item);
			item = fieldUpdater.merge(item);
			if (item.additional == null)
				syncToClient(item);
			else
				item.additional.forEach(i -> syncToClient(i));
			return item;
		} finally {
			fieldUpdater.clear();
		}
	}

	private DataItem syncToClient(DataItem item) {
		if (protocol == null) {
			protocol = new SyncViewToClients();
			protocol.providerid = getProviderId();
			protocol.sessionids.addAll(getSessionIds());
			protocol.classindex = getClassIndex();
			protocol.instanceindex = getInstanceIndex();
			protocol.synctype = getSyncDataType();
			sstream = prepareStringHeader(protocol);
			delaySchedule(() -> hashSchedule(() -> flush()));
		}
		protocol.vardatas.add(item.data);
		if (sstream != null)
			sstream.append(item.text);
		return item;
	}

	void flush() {
		if (!(protocol == null || protocol.sessionids.isEmpty() || protocol.vardatas.isEmpty() || isClosed())) {
			assembleProtocolData(protocol, sstream);
			try {
				getViewContext().syncViewToClients(protocol);
			} catch (Throwable e) {
				if (Trace.isErrorEnabled())
					Trace.error("syncToClient view = " + AutoView.this + " sessions = " + protocol.sessionids, e);
			}
		}
		protocol = null;
	}
}
