package limax.provider;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import limax.codec.Octets;
import limax.codec.OctetsStream;
import limax.codec.StringStream;
import limax.provider.providerendpoint.SyncViewToClients;
import limax.providerendpoint.ViewVariableData;
import limax.util.Resource;

public abstract class View {
	static class DataItem {
		final ViewVariableData data;
		final String text;
		final List<DataItem> additional;

		DataItem(ViewVariableData data, String text) {
			this(data, text, null);
		}

		DataItem(ViewVariableData data, String text, List<DataItem> additional) {
			this.data = data;
			this.text = text;
			this.additional = additional;
		}
	}

	private final ViewContextImpl context;
	private final ViewStub stub;
	private final Resource resource;
	private final DataItem[] history;
	final static Octets NullData = new Octets(0);
	private boolean closed = false;

	protected abstract void processControl(byte controlIndex, OctetsStream controlparameter, long sessionid);

	protected abstract void onMessage(String message, long sessionid);

	protected abstract void onClose();

	public abstract void hashSchedule(Runnable task);

	View(ViewContext context, ViewStub stub, Resource resource, int nFields) {
		this.context = (ViewContextImpl) context;
		this.stub = stub;
		this.resource = resource;
		this.history = new DataItem[nFields];
	}

	short getClassIndex() {
		return stub.getClassIndex();
	}

	short getProviderId() {
		return context.getProviderId();
	}

	protected boolean isScriptEnabled() {
		return context.isScriptEnabled();
	}

	final ViewContextImpl getViewContext() {
		return context;
	}

	synchronized void close() {
		closed = true;
		resource.close();
		onClose();
		getViewContext().onClosed(this);
	}

	protected final boolean isClosed() {
		return closed;
	}

	String getTouch(byte varindex) {
		return null;
	}

	protected final void update(byte varindex, Octets vardata, String strdata) {
		hashSchedule(() -> syncToClient(varindex, vardata, strdata));
	}

	Collection<DataItem> getRecentDataItem(Collection<Byte> varindexes) {
		return varindexes.stream().map(i -> history[i]).filter(Objects::nonNull).collect(Collectors.toList());
	}

	Collection<DataItem> getRecentDataItem() {
		return Arrays.stream(history).filter(Objects::nonNull).collect(Collectors.toList());
	}

	StringStream prepareStringHeader(SyncViewToClients protocol) {
		return isScriptEnabled() ? new StringStream(context.getDataDictionary()).marshal(protocol.providerid)
				.marshal(protocol.classindex).marshal(protocol.instanceindex).marshal(protocol.synctype).append("P")
				: null;
	}

	DataItem syncToClient(byte varindex, Octets vardata, String strdata) {
		boolean removed = vardata.size() == 0;
		DataItem item = history[varindex];
		if (item != null) {
			if (removed) {
				item = new DataItem(new ViewVariableData((byte) (varindex | 0x80), (byte) -1, vardata, NullData),
						strdata);
				history[varindex] = null;
			} else if (vardata.equals(item.data.data)) {
				item = new DataItem(new ViewVariableData(varindex, (byte) -1, NullData, NullData), getTouch(varindex));
			} else {
				item = new DataItem(new ViewVariableData(varindex, (byte) -1, vardata.fit(), NullData), strdata);
				history[varindex] = item;
			}
		} else {
			if (removed) {
				item = null;
			} else {
				item = new DataItem(new ViewVariableData(varindex, (byte) -1, vardata.fit(), NullData), strdata);
				history[varindex] = item;
			}
		}
		return item;
	}

	void scheduleProcessControl(byte controlIndex, OctetsStream controlparameter, long sessionid) {
		hashSchedule(() -> {
			if (!closed)
				processControl(controlIndex, controlparameter, sessionid);
		});
	}

	void scheduleOnMessage(String message, long sessionid) {
		hashSchedule(() -> {
			if (!closed)
				onMessage(message, sessionid);
		});
	}

	@Override
	public String toString() {
		return "[ class = " + getClass().getName() + " ProviderId = " + getProviderId() + " classindex = "
				+ getClassIndex() + "]";
	}
}
