package limax.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class JSON {
	private final Object data;

	public static final Object NULL = new Object() {
		@Override
		public String toString() {
			return "null";
		}
	};

	public class Exception extends java.lang.Exception {
		private static final long serialVersionUID = 7067191163493197089L;

		Exception(String s, int p) {
			super("string-length = " + s.length() + " parse error at " + p);
		}

		Exception(String message, Throwable e) {
			super(message, e);
		}

		Exception(Throwable e) {
			super(e);
		}
	}

	private static class Parser {
		private final String s;
		private int p;

		Parser(String s) {
			this.s = s;
			this.p = 0;
		}

		char get() {
			char ch = s.charAt(p);
			while (Character.isWhitespace(ch))
				ch = s.charAt(++p);
			return ch;
		}

		Object parseObject() {
			Map<String, Object> map = new HashMap<String, Object>();
			p++;
			while (true) {
				if (get() == '}') {
					p++;
					return map;
				}
				String key = (String) parseValue();
				char ch = get();
				if (ch != ':' && ch != '=')
					throw new RuntimeException();
				p++;
				map.put(key, parseValue());
				switch (get()) {
				case '}':
					p++;
					return map;
				case ',':
				case ';':
					p++;
					break;
				default:
					throw new RuntimeException();
				}
			}
		}

		List<Object> parseList() {
			List<Object> c = new ArrayList<>();
			p++;
			if (get() == ']') {
				p++;
				return c;
			}
			while (true) {
				c.add(parseValue());
				switch (get()) {
				case ']':
					p++;
					return c;
				case ',':
				case ';':
					p++;
					break;
				default:
					throw new RuntimeException();
				}
			}
		}

		String parseString() {
			StringBuilder sb = new StringBuilder();
			while (true) {
				char ch = s.charAt(++p);
				if (ch == '"') {
					p++;
					return sb.toString();
				}
				if (ch == '\\') {
					switch (ch = s.charAt(++p)) {
					case '"':
						sb.append('"');
						break;
					case '\\':
						sb.append('\\');
						break;
					case '/':
						// sb.append('\/'); ?? http://www.json.org
						break;
					case 'b':
						sb.append('\b');
						break;
					case 'f':
						sb.append('\f');
						break;
					case 'n':
						sb.append('\n');
						break;
					case 'r':
						sb.append('\r');
						break;
					case 't':
						sb.append('\t');
						break;
					case 'u':
						int v = Character.digit(s.charAt(++p), 16);
						v = (v << 4) | Character.digit(s.charAt(++p), 16);
						v = (v << 4) | Character.digit(s.charAt(++p), 16);
						v = (v << 4) | Character.digit(s.charAt(++p), 16);
						sb.append((char) v);
						break;
					}
				} else {
					sb.append(ch);
				}
			}
		}

		Object parseNumber() {
			StringBuilder sb = new StringBuilder();
			sb.append(s.charAt(p));
			while (true) {
				if (++p == s.length())
					return sb.toString();
				char ch = s.charAt(p);
				if (ch == ']' || ch == '}' || ch == ',' || Character.isWhitespace(ch))
					return sb.toString();
				sb.append(ch);
			}
		}

		Object parseValue() {
			char ch = get();
			switch (ch) {
			case '{':
				return parseObject();
			case '[':
				return parseList();
			case '"':
				return parseString();
			case '-':
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				return parseNumber();
			case 't':
				if (s.charAt(++p) != 'r' || s.charAt(++p) != 'u' || s.charAt(++p) != 'e')
					throw new RuntimeException();
				p++;
				return Boolean.TRUE;
			case 'f':
				if (s.charAt(++p) != 'a' || s.charAt(++p) != 'l' || s.charAt(++p) != 's' || s.charAt(++p) != 'e')
					throw new RuntimeException();
				p++;
				return Boolean.FALSE;
			case 'n':
				if (s.charAt(++p) != 'u' || s.charAt(++p) != 'l' || s.charAt(++p) != 'l')
					throw new RuntimeException();
				p++;
				return JSON.NULL;
			}
			throw new RuntimeException();
		}
	}

	private JSON(String text) throws Exception {
		Parser parser = new Parser(text);
		try {
			this.data = parser.parseValue();
		} catch (java.lang.Exception e) {
			throw new Exception(parser.s, parser.p);
		}
	}

	private JSON(Object data) {
		this.data = data;
	}

	@Override
	public String toString() {
		return data.toString();
	}

	@SuppressWarnings("unchecked")
	public JSON get(String key) throws Exception {
		try {
			return new JSON(((Map<String, Object>) data).get(key));
		} catch (Throwable t) {
			throw new Exception(t);
		}
	}

	@SuppressWarnings("unchecked")
	public Set<String> keySet() throws Exception {
		try {
			return ((Map<String, Object>) data).keySet();
		} catch (Throwable t) {
			throw new Exception(t);
		}
	}

	@SuppressWarnings("unchecked")
	public JSON get(int index) throws Exception {
		try {
			return new JSON(((List<Object>) data).get(index));
		} catch (Throwable t) {
			throw new Exception(t);
		}
	}

	@SuppressWarnings("unchecked")
	JSON[] toArray() throws Exception {
		try {
			List<JSON> list = new ArrayList<>();
			for (Object o : (List<Object>) data)
				list.add(new JSON(o));
			return list.toArray(new JSON[0]);
		} catch (Throwable t) {
			throw new Exception(t);
		}
	}

	public boolean booleanValue() throws Exception {
		try {
			if (data instanceof Boolean)
				return ((Boolean) data).booleanValue();
			return Boolean.parseBoolean((String) data);
		} catch (Throwable t) {
			throw new Exception(data + " cast to boolean", t);
		}
	}

	public int intValue() throws Exception {
		try {
			if (data instanceof Boolean)
				return ((Boolean) data).booleanValue() ? 1 : 0;
			return Integer.parseInt((String) data);
		} catch (Throwable t) {
			throw new Exception(data + " cast to int", t);
		}
	}

	public long longValue() throws Exception {
		try {
			if (data instanceof Boolean)
				return ((Boolean) data).booleanValue() ? 1l : 0;
			return Long.parseLong((String) data);
		} catch (Throwable t) {
			throw new Exception(data + " cast to long", t);
		}
	}

	public double doubleValue() throws Exception {
		try {
			if (data instanceof Boolean)
				return ((Boolean) data).booleanValue() ? 1.0 : 0.0;
			return Double.parseDouble((String) data);
		} catch (Throwable t) {
			throw new Exception(data + " cast to double", t);
		}
	}

	public boolean isNull() {
		return data == JSON.NULL;
	}

	public static JSON parse(String text) throws Exception {
		return new JSON(text);
	}
}
